"""This package provides Sparkle's wrappers for Pandas DataFrames."""
from sparkle.structures.feature_dataframe import FeatureDataFrame
from sparkle.structures.performance_dataframe import PerformanceDataFrame
